import xml.etree.ElementTree as xml_element_tree
import datetime
import hashlib
import hmac
import requests     # pip install requests OR pip3 install requests
import xml.dom.minidom


# Constants
SERVICE = 's3'
CONTENT_TYPE = 'text/xml'
REQUEST_PARAMETERS = 'compliance='
ALGORITHM = 'AWS4-HMAC-SHA256'


def get_available_regions() -> dict:
    """
    Function to print the list of available Wasabi regions
    :return: Returns the dictionary with region and region names
    """
    # URL to get list of available region in Wasabi
    url = 'https://s3.wasabisys.com/?describeRegions'

    # Capture the response dynamically from the URL
    desc_regions_response = requests.get(url)

    # Dictionary to capture the regions
    region_dict = {}

    # Loop through the XML element and write data to the region dictionary
    region_counter = 0
    for child in xml_element_tree.fromstring(desc_regions_response.text):
        region_dict[region_counter] = {'region': child[0].text, 'region_name': child[1].text}
        region_counter += 1

    print("********* Available Regions *********")

    # Parse the dictionary and print the list of available regions
    for region in region_dict:
        print(
            f'{region_dict[region]["region_name"].ljust(20, " ")} | {region_dict[region]["region"]}'
        )

    return region_dict


# Key derivation functions. See:
# http://docs.aws.amazon.com/general/latest/gr/signature-v4-examples.html#signature-v4-examples-python
def sign(key, msg):
    return hmac.new(key, msg.encode('utf-8'), hashlib.sha256).digest()


def getSignatureKey(key, dateStamp, regionName, serviceName):
    kDate = sign(('AWS4' + key).encode('utf-8'), dateStamp)
    kRegion = sign(kDate, regionName)
    kService = sign(kRegion, serviceName)
    kSigning = sign(kService, 'aws4_request')
    return kSigning


def wasabi_compliance(method: str, wasabi_host: str, endpoint: str, dar_flag: str = "N", retention_time_in_days: int = -1) -> str:
    """
    Function to display the compliance setting for a bucket. We can also modify delete after
    retention settings & retention time (in days) for the bucket. For both GET/PUT methods, we need to
    sign the requests and this function covers the signing methods for both GET and PUT API calls.

    :param method: GET or PUT methods
    :param wasabi_host: Wasabi Host Name
    :param endpoint: Wasabi End Point based on the host name
    :param dar_flag: Delete After Retention Flag `Y` to set it as `true` or `N` to set it as `false`
    :param retention_time_in_days: Retention Time in days, default is -1 if the customer do not wish to modify
    :return: `disabled` if compliance is `off`. `None` if compliance is enabled or compliance is updated
    """

    # Create a date for headers and the credential string
    t = datetime.datetime.utcnow()
    amzdate = t.strftime('%Y%m%dT%H%M%SZ')
    datestamp = t.strftime('%Y%m%d')  # Date w/o time, used in credential scope

    # Create canonical URI--the part of the URI from domain to query
    canonical_uri = f'/{wasabi_bucket_name}'

    # Create the canonical query string
    canonical_querystring = REQUEST_PARAMETERS

    if method == "GET":

        # Create the canonical headers and signed headers
        canonical_headers = 'host:' + wasabi_host + '\n' + 'x-amz-date:' + amzdate + '\n'

        # Create the list of signed headers
        signed_headers = 'host;x-amz-date'

        # Create payload hash
        payload_hash = hashlib.sha256(''.encode('utf-8')).hexdigest()

        # Combine elements to create canonical request
        canonical_request = method + '\n' + canonical_uri + '\n' + canonical_querystring + '\n' + canonical_headers + '\n' + signed_headers + '\n' + payload_hash

        # Credential scope
        credential_scope = datestamp + '/' + wasabi_region + '/' + SERVICE + '/' + 'aws4_request'

        # Create the string to sign
        string_to_sign = ALGORITHM + '\n' + amzdate + '\n' + credential_scope + '\n' + hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()

        # Calculate the Signature
        # Create the signing key using the function defined above.
        signing_key = getSignatureKey(wasabi_secret_access_key, datestamp, wasabi_region, SERVICE)

        # Sign the string_to_sign using the signing_key
        signature = hmac.new(signing_key, string_to_sign.encode('utf-8'), hashlib.sha256).hexdigest()

        # Add signing information to the request
        authorization_header = ALGORITHM + ' ' + 'Credential=' + wasabi_access_key + '/' + credential_scope + ', ' + 'SignedHeaders=' + signed_headers + ', ' + 'Signature=' + signature

        # Create header for request
        headers = {'x-amz-date': amzdate, 'Authorization': authorization_header, 'x-amz-content-sha256': payload_hash}

        # Send the request
        request_url = endpoint + '?compliance'
        get_response = requests.get(request_url, headers=headers)

        # Display the current settings
        print()

        if get_response.status_code != 200:
            get_xml_out = xml.dom.minidom.parseString(get_response.text)
            print()
            print(get_xml_out.toprettyxml())
            # print(get_response.text)
            exit()    # Exit the program if there is an issue
        else:  # get_response.status_code == 200
            output_dict = {}
            for index, child in enumerate(xml_element_tree.fromstring(get_response.content)):
                output_dict[index] = child.text

            print("Compliance settings for Bucket: ", wasabi_bucket_name)
            print("Compliance Mode: ", output_dict[0])
            # We will get the below information only if Compliance is Enabled
            if output_dict[0] == 'enabled':
                print("Compliance Lock: ", output_dict[2])
                print("Compliance Lock Time: ", output_dict[1])
                print("Retention Time (in days): ", output_dict[3])
                if output_dict[4] == 'true':    # Display Conditional Hold only if it has been enabled
                    print("Conditional Hold: ", output_dict[4])
                print("Delete After Retention: ", output_dict[5])
                print()
            else:   # Status of compliance "disabled"
                return output_dict[0]

    else:   # method == "PUT"

        # Set delete after retention setting
        if dar_flag == "Y":
            dar_setting = "true"
        else:   # flag == "N"
            dar_setting = "false"

        # Initialize payload
        payload = ""

        if retention_time_in_days == -1:
            # Set the payload to only modify DAR
            payload = f"""
                <?xml version="1.0" encoding="UTF-8"?>
                <BucketComplianceConfiguration xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                    <DeleteAfterRetention>{dar_setting}</DeleteAfterRetention>
                </BucketComplianceConfiguration>
                """
        elif retention_time_in_days >= 0:
            # Set the payload to modify both DAR & Retention days
            payload = f"""
                <?xml version="1.0" encoding="UTF-8"?>
                <BucketComplianceConfiguration xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                    <DeleteAfterRetention>{dar_setting}</DeleteAfterRetention>
                    <RetentionDays>{retention_time_in_days}</RetentionDays>
                </BucketComplianceConfiguration>
                """

        # Create payload hash
        payload_hash = hashlib.sha256(payload.encode('utf-8')).hexdigest()

        # Create Canonical Headers
        canonical_headers = 'host:' + wasabi_host + '\n' + 'content-type:' + CONTENT_TYPE + '\n' + 'x-amz-content-sha256:' + payload_hash + '\n' + 'x-amz-date:' + amzdate + '\n'

        # Create the list of signed headers
        signed_headers = 'host;content-type;x-amz-content-sha256;x-amz-date'

        # Combine elements to create canonical request
        canonical_request = method + '\n' + canonical_uri + '\n' + canonical_querystring + '\n' + canonical_headers + '\n' + signed_headers + '\n' + payload_hash

        # Credential scope
        credential_scope = datestamp + '/' + wasabi_region + '/' + SERVICE + '/' + 'aws4_request'

        # Create the string to sign
        string_to_sign = ALGORITHM + '\n' + amzdate + '\n' + credential_scope + '\n' + hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()

        # Calculate the signature
        signing_key = getSignatureKey(wasabi_secret_access_key, datestamp, wasabi_region, SERVICE)

        # Sign the string_to_sign using the signing_key
        signature = hmac.new(signing_key, string_to_sign.encode('utf-8'), hashlib.sha256).hexdigest()

        # Add signing information to the request
        authorization_header = ALGORITHM + ' ' + 'Credential=' + wasabi_access_key + '/' + credential_scope + ', ' + 'SignedHeaders=' + signed_headers + ', ' + 'Signature=' + signature

        # Create header for request
        headers = {'Content-Type': CONTENT_TYPE, 'x-amz-date': amzdate, 'Authorization': authorization_header, 'x-amz-content-sha256': payload_hash}

        # Send the request
        request_url = endpoint + '?compliance'
        put_response = requests.put(request_url, data=payload, headers=headers)

        if put_response.status_code != 200:
            put_xml_out = xml.dom.minidom.parseString(put_response.text)
            print()
            print(put_xml_out.toprettyxml())
            # print(put_response.text)
            exit()    # Exit the program if there is an issue
        else:  # put_response.status_code == 200
            print()
            print(f"Delete after retention has been set to `{dar_setting}` for bucket `{wasabi_bucket_name}`")


# Display the list of Wasabi regions
av_reg_dict = get_available_regions()

# Get input from the customer...
print()
wasabi_access_key = input("Please enter Wasabi Access Key: ") or None
wasabi_secret_access_key = input("Please enter Wasabi Secret Access Key: ") or None
wasabi_bucket_name = input("Please enter Wasabi bucket name: ")
wasabi_region = input("Please enter the region (Press ENTER for Default region : us-east-1): ")

# Program should terminate if the Access Key or Secret Access key is unavailable
if wasabi_access_key is None or wasabi_secret_access_key is None:
    print()
    print('No access key is available.')
else:
    # Validate Wasabi Region
    counter = 0
    for regions in av_reg_dict:
        if wasabi_region in av_reg_dict[regions]["region"]:
            counter += 1

    if counter == 0:
        print()
        print("Invalid Wasabi Region: {}".format(wasabi_region))
        exit()  # Exit the program if the region is Invalid

    # Set the Default Region
    if wasabi_region == "":
        wasabi_region += "us-east-1"

    # Set Wasabi Host and Wasabi endpoint based on customers input
    wasabi_host_name = f's3.{wasabi_region}.wasabisys.com'
    wasabi_endpoint = f'https://{wasabi_host_name}/{wasabi_bucket_name}'

    # Display current Compliance Setting for the bucket
    compliance_status = wasabi_compliance('GET', wasabi_host_name, wasabi_endpoint)

    if compliance_status != 'disabled':
        # Loop to modify delete after retention & Retention Days setting for a bucket
        while True:

            customer_input = input("""Please select 
                "Y" to enable delete after retention
                "N" to disable delete after retention
                "Q" to quit: """)

            if customer_input == "Y":  # Enable delete after retention
                retention_days = int(input("Please set Retention Time (in days). Press ENTER if you do not wish to modify: ") or -1)    # Default value is -1 if the customer do not wish to modify retention days
                if retention_days == 0:
                    print("Retention Time (in days) should be greater than ZERO. Please try again.")
                    print()
                    continue
                wasabi_compliance('PUT', wasabi_host_name, wasabi_endpoint, customer_input, retention_days)
            elif customer_input == "N":  # Disable delete after retention
                wasabi_compliance('PUT', wasabi_host_name, wasabi_endpoint)
            elif customer_input == "Q":
                break

            # Display the compliance settings after the change
            wasabi_compliance('GET', wasabi_host_name, wasabi_endpoint)

print()
print("Program Ended...")
